# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys


class StubPackage(object):
    pass


# load all tests from /elements/*/tests/ dirs.
# conceptually load_tests should be in __init__, but see
# http://bugs.python.org/issue16662 instead. So, its here in test_elements.py
def load_tests(loader, tests, pattern):
    """load tests for diskimage_builder elements."""
    if pattern is None:
        # http://bugs.python.org/issue11218
        pattern = "test*.py"
    this_dir = os.path.dirname(__file__)
    elements_dir = os.path.join(this_dir, "..", "..", "elements")
    # Make a fake elements top level package, as discovery doesn't let us
    # override the python path.
    package = StubPackage()
    package.__path__ = [elements_dir]
    sys.modules['elements'] = package
    elements = os.listdir(elements_dir)
    for element in elements:
        element_dir = os.path.join(elements_dir, element)
        tests_path = os.path.join(element_dir, "tests")
        if (not os.path.isdir(tests_path) or
                not os.path.isfile(os.path.join(tests_path, '__init__.py'))):
            continue
        # Create a 'package' for the element, so it can be imported.
        package = StubPackage()
        package.__path__ = [element_dir]
        sys.modules['elements.%s' % element] = package
        # Try importing the test module
        package_tests = loader.discover(start_dir=tests_path, pattern=pattern)
        tests.addTests(package_tests)
    return tests



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
import argparse
import collections
import os
import sys


def get_elements_dir():
    if not os.environ.get('ELEMENTS_PATH'):
        raise Exception("$ELEMENTS_PATH must be set.")
    return os.environ['ELEMENTS_PATH']


def _get_set(element, fname, elements_dir=None):
    if elements_dir is None:
        elements_dir = get_elements_dir()

    for path in elements_dir.split(':'):
        element_deps_path = (os.path.join(path, element, fname))
        try:
            with open(element_deps_path) as element_deps:
                return set([line.strip() for line in element_deps])
        except IOError as e:
            if os.path.exists(os.path.join(path, element)) and e.errno == 2:
                return set()
            if e.errno == 2:
                continue
            else:
                raise

    sys.stderr.write("ERROR: Element '%s' not found in '%s'\n" %
                     (element, elements_dir))
    sys.exit(-1)


def provides(element, elements_dir=None):
    """Return the set of elements provided by the specified element.

    :param element: name of a single element
    :param elements_dir: the elements dir to read from. If not supplied,
                         inferred by calling get_elements_dir().

    :return: a set just containing all elements that the specified element
             provides.
    """
    return _get_set(element, 'element-provides', elements_dir)


def dependencies(element, elements_dir=None):
    """Return the non-transitive set of dependencies for a single element.

    :param element: name of a single element
    :param elements_dir: the elements dir to read from. If not supplied,
                         inferred by calling get_elements_dir().

    :return: a set just containing all elements that the specified element
             depends on.
    """
    return _get_set(element, 'element-deps', elements_dir)


def expand_dependencies(user_elements, elements_dir=None):
    """Expand user requested elements using element-deps files.

    Arguments:
    :param user_elements: iterable enumerating the elements a user requested
    :param elements_dir: the elements dir to read from. Passed directly to
                         dependencies()

    :return: a set containing user_elements and all dependent elements
             including any transitive dependencies.
    """
    final_elements = set(user_elements)
    check_queue = collections.deque(user_elements)
    provided = set()

    while check_queue:
        # bug #1303911 - run through the provided elements first to avoid
        # adding unwanted dependencies and looking for virtual elements
        element = check_queue.popleft()
        if element in provided:
            continue
        deps = dependencies(element, elements_dir)
        provided.update(provides(element, elements_dir))
        check_queue.extend(deps - (final_elements | provided))
        final_elements.update(deps)

    if "operating-system" not in provided:
        sys.stderr.write(
            "ERROR: Please include an operating system element.\n")
        sys.exit(-1)

    conflicts = set(user_elements) & provided
    if conflicts:
        sys.stderr.write("ERROR: Following elements were explicitly required "
                         "but are provided by other included elements: %s\n" %
                         ", ".join(conflicts))
        sys.exit(-1)
    return final_elements - provided


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('elements', nargs='+',
                        help='display dependencies of the given elements')
    parser.add_argument('--expand-dependencies', '-d', action='store_true',
                        default=False,
                        help=('(DEPRECATED) print expanded dependencies '
                              'of all args'))

    args = parser.parse_args(argv[1:])

    if args.expand_dependencies:
        print("WARNING: expand-dependencies flag is deprecated,  "
              "and is now on by default.", file=sys.stderr)

    print(' '.join(expand_dependencies(args.elements)))
    return 0



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import fixtures
import testtools

from diskimage_builder import element_dependencies

data_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'test-elements'))


def _populate_element(element_dir, element_name, element_deps=[], provides=[]):
        element_home = os.path.join(element_dir, element_name)
        os.mkdir(element_home)
        deps_path = os.path.join(element_home, 'element-deps')
        with open(deps_path, 'w') as deps_file:
            deps_file.write("\n".join(element_deps))
        provides_path = os.path.join(element_home, 'element-provides')
        with open(provides_path, 'w') as provides_file:
            provides_file.write("\n".join(provides))


class TestElementDeps(testtools.TestCase):

    def setUp(self):
        super(TestElementDeps, self).setUp()
        self.element_dir = self.useFixture(fixtures.TempDir()).path
        _populate_element(self.element_dir, 'requires-foo', ['foo'])
        _populate_element(self.element_dir,
                          'foo',
                          [],
                          ['operating-system'])
        _populate_element(self.element_dir,
                          'requires-requires-foo',
                          ['requires-foo'])
        _populate_element(self.element_dir, 'self', ['self'])
        _populate_element(self.element_dir,
                          'provides_virtual',
                          [],
                          ['virtual'])
        _populate_element(self.element_dir,
                          'requires_virtual',
                          ['virtual'],
                          ['operating-system'])
        _populate_element(self.element_dir, 'virtual', ['extra_dependency'])
        _populate_element(self.element_dir, 'extra_dependency', [])
        _populate_element(self.element_dir,
                          'circular1',
                          ['circular2'],
                          ['operating-system'])
        _populate_element(self.element_dir, 'circular2', ['circular1'])
        _populate_element(self.element_dir,
                          'provides_new_virtual',
                          [],
                          ['new_virtual', 'operating-system'])
        _populate_element(self.element_dir,
                          'requires_new_virtual',
                          ['new_virtual'])

    def test_non_transitive_deps(self):
        result = element_dependencies.expand_dependencies(
            ['requires-foo'],
            elements_dir=self.element_dir)
        self.assertEqual(set(['requires-foo', 'foo']), result)

    def test_missing_deps(self):
        self.assertRaises(SystemExit,
                          element_dependencies.expand_dependencies, ['fake'],
                          self.element_dir)

    def test_transitive_deps(self):
        result = element_dependencies.expand_dependencies(
            ['requires-requires-foo'], elements_dir=self.element_dir)
        self.assertEqual(set(['requires-requires-foo',
                              'requires-foo',
                              'foo']), result)

    def test_no_deps(self):
        result = element_dependencies.expand_dependencies(
            ['foo'], elements_dir=self.element_dir)
        self.assertEqual(set(['foo']), result)

    def test_self(self):
        result = element_dependencies.expand_dependencies(
            ['self', 'foo'], elements_dir=self.element_dir)
        self.assertEqual(set(['self', 'foo']), result)

    def test_circular(self):
        result = element_dependencies.expand_dependencies(
            ['circular1'], elements_dir=self.element_dir)
        self.assertEqual(set(['circular1', 'circular2']), result)

    def test_provide(self):
        result = element_dependencies.expand_dependencies(
            ['provides_virtual', 'requires_virtual'],
            elements_dir=self.element_dir)
        self.assertEqual(set(['requires_virtual', 'provides_virtual']), result)

    def test_provide_conflict(self):
        self.assertRaises(SystemExit,
                          element_dependencies.expand_dependencies,
                          ['virtual', 'provides_virtual'],
                          self.element_dir)

    def test_provide_virtual_ordering(self):
        result = element_dependencies.expand_dependencies(
            ['requires_new_virtual', 'provides_new_virtual'],
            elements_dir=self.element_dir)
        self.assertEqual(set(['requires_new_virtual', 'provides_new_virtual']),
                         result)

    def test_no_os_element(self):
        self.assertRaises(SystemExit,
                          element_dependencies.expand_dependencies,
                          ['provides_virtual'],
                          elements_dir=self.element_dir)

    def test_duplicated_os_passed_as_element(self):
        self.assertRaises(SystemExit,
                          element_dependencies.expand_dependencies,
                          ['circular1', 'operating-system'],
                          elements_dir=self.element_dir)


class TestElements(testtools.TestCase):
    def test_depends_on_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable('ELEMENTS_PATH', '/foo/bar'))
        self.assertEqual('/foo/bar', element_dependencies.get_elements_dir())

    def test_env_not_set(self):
        self.useFixture(fixtures.EnvironmentVariable('ELEMENTS_PATH', ''))
        self.assertRaises(Exception, element_dependencies.get_elements_dir, ())
